/*

=head1 NAME

tinysh2 -- a tiny shell with a minimal set of features

=head1 SYNOPSIS

B<tinysh2>

=head1 DESCRIPTION

B<tinysh2> reads command lines from its standard input
and executes them. ``tinysh2> '' is given as prompt.

Each command line consists of space-separated tokens. Special tokens
begin with ``<'', indicating the input file, ``>'', indicating the output
file, ``>>'', specifying an output file that is to be extended, and
``&'', indicating a background execution. The first non-special token
specifies the command name which must be either an absolute path of an
executable file or locatable within the list of directories provided by
the environment variable B<PATH>.

=head1 DIAGNOSTICS

B<tinysh2> prints the exit code of a terminated program if it is non-zero.
An exit code of 255 is used by subprocesses if they are unable to start
the command for some reason.

=head1 BUGS

No pipelines, no builtins, no shell variables.

=head1 AUTHOR

Andreas Borchert

=cut

*/

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include "command.h"
#include "sareadline.h"
#include "strlist.h"
#include "tokenizer.h"

static volatile sig_atomic_t interrupted = 0;

void interrupt_handler(int sig) {
   interrupted = 1;
}

void print_child_status(pid_t pid, int wstat) {
   printf("[%d] ", (int) pid);
   if (WIFEXITED(wstat)) {
      printf("exit %d", WEXITSTATUS(wstat));
   } else if (WIFSIGNALED(wstat)) {
      printf("terminated with signal %d", WTERMSIG(wstat));
      if (WCOREDUMP(wstat)) printf(" (core dump)");
   } else {
      printf("???");
   }
   printf("\n");
}

void status_report(void) {
   pid_t pid; int wstat;
   while ((pid = waitpid((pid_t)-1, &wstat, WNOHANG)) > 0) {
      print_child_status(pid, wstat);
   }
}

int getline(stralloc* line) {
   int first = 1;
   interrupted = 0;
   for(;;) {
      if (interrupted) {
         interrupted = 0;
         printf("\n");
         first = 1;
      }
      if (first) {
         status_report();
         printf("%% ");
         first = 0;
      }
      errno = 0;
      if (readline(stdin, line)) return 1;
      if (errno != EINTR) return 0;
   }
}

int main() {
   struct sigaction action = {0};
   action.sa_handler = interrupt_handler;
   if (sigaction(SIGINT, &action, 0) != 0 ||
         sigaction(SIGQUIT, &action, 0) != 0) {
      perror("sigaction");
   }

   stralloc line = {0};
   while (getline(&line)) {
      strlist tokens = {0};
      stralloc_0(&line); /* required by tokenizer() */
      if (!tokenizer(&line, &tokens)) break;
      if (tokens.len == 0) continue;
      command cmd = {0};
      if (!scan_command(&tokens, &cmd)) continue;

      sighold(SIGINT); sighold(SIGQUIT);
      pid_t child = fork();
      if (child == -1) {
         perror("fork"); continue;
      }
      if (child == 0) {
         sigrelse(SIGINT); sigrelse(SIGQUIT);
         if (cmd.background) {
            sigignore(SIGINT); sigignore(SIGQUIT);
         }
         exec_command(&cmd);
         perror(cmd.cmdname);
         exit(255);
      }

      if (cmd.background) {
         printf("%d\n", (int)child);
      } else {
         int wstat;
         pid_t pid = waitpid(child, &wstat, 0);
         if (!WIFEXITED(wstat) || WEXITSTATUS(wstat)) {
            print_child_status(pid, wstat);
         }
      }
      sigrelse(SIGINT); sigrelse(SIGQUIT);
   }
}
